@@ -88,8 +88,6 @@ def fastjet_clusters(
 
     Notes
     -----
-    This is a wrapper around FastJet's implementation.
-
     ``p_val`` set to ``-1`` gives **anti-kT**, ``0`` gives
     **Cambridge-Aachen**, and ``1`` gives **kT** clusterings.
     """
@@ -133,7 +131,8 @@ def find_vertex(
     Parameters
     ----------
     graph : Graphicle
-        Graphicle object, which must contain at least edge and pdg data.
+        ``Graphicle`` object, which must contain at least edge and pdg
+        data.
     pdgs_in : set of ints
         Subset of pdg codes to match against the incoming particles.
     pdgs_out : set of ints
@@ -141,9 +140,14 @@ def find_vertex(
 
     Returns
     -------
-    vertices : array of ints
+    vertices : ndarray[int32]
         List the vertex ids which match the passed incoming and outgoing
         pdg codes.
+
+    Raises
+    ------
+    ValueError
+        Raised if ``pdgs_in`` and ``pdgs_out`` are both left blank.
     """
     # preparing the search sets
     search = dict()
@@ -373,8 +377,8 @@ def partition_descendants(
     for vtx_id, pcls_in in hadron_vtxs:
         vtx_desc = vertex_descendants(graph.adj, vtx_id)
         for name, branch in hier.items():
-            for _, mask in _leaf_mask_iter(branch, name):
-                if vtx_id not in graph.edges["out"][mask.data]:
+            for _, mask in _leaf_mask_iter(name, branch):
+                if vtx_id not in graph.edges["out"][mask]:
                     continue
                 mask.data = _partition_vertex(
                     mask,
@@ -387,10 +391,24 @@ def partition_descendants(
     return hier
 
 
-def _hard_matrix(hard_desc: gcl.MaskGroup) -> base.BoolVector:
-    """Returns a square adjacency matrix for the hard process. Rows
-    represent each hard parton, and the values indicate whether another
-    given hard parton is a descendant of it.
+def _hard_ancestor_matrix(hard_desc: gcl.MaskGroup) -> base.BoolVector:
+    """Returns a directed adjacency matrix for the hard process.
+
+    Parameters
+    ----------
+    hard_desc : MaskGroup
+        Collection of masks indicating descendants from hard partons,
+        over the hard process only. Obtain with ``hard_descendants()``
+        and then applying ``StatusArray.hard_mask`` (with ``'incoming'``
+        removed).
+
+    Returns
+    -------
+    hard_matrix : ndarray[bool_]
+        Square directed adjacency matrix, linking partons in the hard
+        process to their descendants. Rows represent each hard parton,
+        and columns indicate which hard partons descend from it
+        (including descendants-of-descendants).
     """
     num_hard = len(hard_desc)
     mat = np.zeros((num_hard, num_hard), dtype="<?")
@@ -399,39 +417,108 @@ def _hard_matrix(hard_desc: gcl.MaskGroup) -> base.BoolVector:
     return mat
 
 
-def _composite_matrix(hard_matrix: base.BoolVector) -> base.BoolVector:
-    """Similar to ``_hard_matrix()``, but columns represent direct
-    children, not grand-children further nested composites.
+def _hard_parent_matrix(hard_matrix: base.BoolVector) -> base.BoolVector:
+    """Returns an adjacency matrix over the hard process of parents to
+    their direct children. Filters out grandchildren *etc*.
+
+    Parameters
+    ----------
+    hard_matrix : ndarray[bool_]
+        Directed adjacency matrix representing the links from ancestors
+        to descendants (including descendants-of-descendants *etc*.)
+        Obtain with ``_hard_ancestor_matrix()``.
+
+    Returns
+    -------
+    hard_parent_matrix : ndarray[bool_]
+        Directed adjacency matrix representing links from parents to
+        direct (first generation) children. Same as ``hard_matrix``, but
+        with grandchildren *etc.* removed.
+
+    Notes
+    -----
+    This function has side-effects on the input. In fact, the return
+    value is the same object in memory as ``hard_matrix``, so could
+    technically be discarded entirely. If this behaviour is problematic,
+    pass ``hard_matrix.copy()`` instead.
     """
-    pcls_with_hard_children = filter(np.any, hard_matrix)
-    asc_parents = sorted(pcls_with_hard_children, key=np.sum)
-    for row_i in reversed(asc_parents):
-        for row_j in asc_parents:
-            if row_i is row_j:
-                continue
-            row_i[row_j] = False
+    pcls_with_hard_parent_dict = filter(np.any, hard_matrix)
+    asc_parents = sorted(pcls_with_hard_parent_dict, key=np.sum)
+    row_pairs = it.product(reversed(asc_parents), asc_parents)
+    for row_i, row_j in filter(lambda pair: op.is_not(*pair), row_pairs):
+        row_i[row_j] = False
     return hard_matrix
 
 
-def _hard_children(
+def _hard_parent_dict(
     comp_mat: base.BoolVector,
     names: ty.Tuple[str, ...],
 ) -> ty.Dict[str, ty.Tuple[str, ...]]:
+    """Creates a mapping between the names of parent partons in the hard
+    process to the names of their children.
+
+    Parameters
+    ----------
+    comp_mat : ndarray[bool_]
+        Directed adjacency matrix for the hard process, with edges from
+        parent hard partons to their immediate children. Each row
+        represents a parent parton, where columns represent the children
+        they are linked to. Obtain this with ``_hard_parent_matrix()``.
+    names : tuple[str, ...]
+        Names of the hard partons, appearing in the same order as the
+        corresponding rows / columns of adjacency matrix representing
+        the hard process (and therefore same order as ``comp_mat``).
+
+    Returns
+    -------
+    parent_dict : dict[str, tuple[str, ...]]
+        Mapping from parents to children in the hard process. Keys are
+        parent names, and values are tuples of children names.
+    """
     parents = filter(lambda parton: np.any(parton[1]), zip(names, comp_mat))
-    comp_dict = dict()
-    for name, parton in parents:
-        comp_dict[name] = tuple(it.compress(names, parton))
-    return comp_dict
+    parents = it.tee(parents, 2)
+    parent_names = map(op.itemgetter(0), parents[0])
+    children_masks = map(op.itemgetter(1), parents[1])
+    children_names = map(it.compress, it.repeat(names), children_masks)
+    children_names = map(tuple, children_names)
+    return dict(zip(parent_names, children_names))
 
 
 def _flat_hierarchy(
     hard_desc: gcl.MaskGroup,
 ) -> ty.Dict[str, ty.Tuple[str, ...]]:
-    """Produces a flat representation of the hard process hierarchy."""
+    """Produces a mapping between partons to their direct children in
+    the hard process. This mapping is flat, *ie.* there is no nesting
+    of dictionaries in dictionaries. Children of one parton may be
+    parents of another, and so may be both included in a child tuple
+    and as a parent key. There is no distinction between grandparents
+    and parents, *etc*.
+
+    Parameters
+    ----------
+    hard_desc : MaskGroup
+        Collection of masks indicating descendants from hard partons,
+        over the hard process only. Obtain with ``hard_descendants()``
+        and then applying ``StatusArray.hard_mask`` (with ``'incoming'``
+        removed).
+
+    Returns
+    -------
+    parent_dict : dict[str, tuple[str, ...]]
+        Mapping from parents to children in the hard process. Keys are
+        parent names, and values are tuples of children names.
+
+    Notes
+    -----
+    In rare cases, when several particles with the same PDG code are
+    present in the hard process, the particle names will be appended
+    with a colon followed by a numerical index, *eg.* ``'b~:0'``. See
+    ``_pdgs_to_keys()`` for more information.
+    """
     names = tuple(hard_desc.keys())
-    hard_matrix = _hard_matrix(hard_desc)
-    comp_matrix = _composite_matrix(hard_matrix)
-    return _hard_children(comp_matrix, names)
+    ancestor_matrix = _hard_ancestor_matrix(hard_desc)
+    parent_matrix = _hard_parent_matrix(ancestor_matrix)
+    return _hard_parent_dict(parent_matrix, names)
 
 
 def _pdgs_to_keys(pdg: gcl.PdgArray) -> ty.Tuple[str, ...]:
@@ -461,6 +548,7 @@ def hard_descendants(
     graph: gcl.Graphicle,
     target: ty.Optional[ty.Iterable[int]] = None,
     sign_sensitive: bool = False,
+    strict: bool = True,
 ) -> gcl.MaskGroup:
     """Returns a ``MaskGroup`` instance to select particle descendants
     of ``target`` hard partons (by PDG code).
@@ -470,7 +558,10 @@ def hard_descendants(
     .. versionadded:: 0.1.0
 
     .. versionchanged:: 0.1.11
-       Target parameter now optional.
+       ``target`` parameter now optional.
+
+    .. versionchanged:: 0.2.5
+       ``strict`` parameter added.
 
     Parameters
     ----------
@@ -487,28 +578,45 @@ def hard_descendants(
         anti-particle partons will be masked, whereas if ``True`` only
         the partons explicitly matching the target sign will be
         considered. Default is ``False``.
+    strict : bool
+        If ``True`` all PDGs in ``target`` must be present in the hard
+        process. If ``False``, only a subset need be present. Default
+        is ``True``.
+
+    Returns
+    -------
+    descendant_masks : MaskGroup
+        Collection of masks over the event indicating the descendants
+        of partons from within the hard process.
+
+    Raises
+    ------
+    ValueError
+        If PDG codes required by ``target`` and ``strict`` are absent
+        from the partons in the hard process.
     """
-    hard_vtxs = list()
-    # get the vertices of the hard partons
-    for stage, mask in graph.hard_mask.items():
-        if stage == "incoming":
-            continue
-        pcls = graph[mask]
-        if target is not None:
-            hard_mask = pcls.pdg.mask(
-                list(target), blacklist=False, sign_sensitive=sign_sensitive
-            )
-            if not np.any(hard_mask):
-                continue
-            pcls = pcls[hard_mask]
-        pdg_keys = _pdgs_to_keys(pcls.pdg)
-        pcl_out_vtxs = map(int, pcls.edges["out"])
-        hard_vtxs.extend(list(zip(pdg_keys, pcl_out_vtxs)))
-    # find the descendants of those vertices
-    masks = gcl.MaskGroup(agg_op="or")
-    for pdg_key, vtx in hard_vtxs:
-        masks[pdg_key] = vertex_descendants(graph.adj, vtx)
-    return masks
+    hard_mask = graph.hard_mask
+    del hard_mask["incoming"]
+    hard_graph = graph[hard_mask]
+    if target is not None:
+        target_pdgs = set(target)
+        target_mask = hard_graph.pdg.mask(
+            target=tuple(target_pdgs),
+            blacklist=False,
+            sign_sensitive=sign_sensitive,
+        )
+        if not np.any(target_mask):
+            raise ValueError("No target PDGs found in hard process.")
+        if (strict is True) and (np.sum(target_mask) != len(target_pdgs)):
+            found_pdgs = set(map(int, hard_graph.pdg[target_mask]))
+            missing_pdgs = target_pdgs - found_pdgs
+            missing_str = ", ".join(map(str, missing_pdgs))
+            raise ValueError(f"Missing PDGs in hard process: {missing_str}.")
+        hard_graph = hard_graph[target_mask]
+    pdg_keys = _pdgs_to_keys(hard_graph.pdg)
+    pcl_out_vtxs = map(int, hard_graph.edges["out"])
+    vtx_descs = map(vertex_descendants, it.repeat(graph.adj), pcl_out_vtxs)
+    return gcl.MaskGroup(cl.OrderedDict(zip(pdg_keys, vtx_descs)), agg_op="or")
 
 
 def hierarchy(
@@ -612,11 +720,14 @@ def hierarchy(
         for key, idx in zip(hard_final_keys, hard_final_idxs):
             desc[key].data[idx] = True
     masks = _make_tree(hard, roots, hard, desc, graph.adj)
-    for root_name, root in masks.items():
-        for name, mask in _leaf_mask_iter(root, root_name, False):
-            if name != "latent":
-                continue
-            mask[mask & graph.hard_mask] = False
+    leaf_iters = it.starmap(
+        fn.partial(_leaf_mask_iter, exclude_latent=False), masks.items()
+    )
+    leaf_iter = it.chain.from_iterable(leaf_iters)
+    leaf_iter = filter(lambda kv: kv[0] == "latent", leaf_iter)
+    leaf_masks = map(op.itemgetter(1), leaf_iter)
+    for leaf_mask in leaf_masks:
+        leaf_mask[leaf_mask & graph.hard_mask] = False
     return masks
 
 
@@ -647,9 +758,7 @@ def _make_tree(
     """
     branch = gcl.MaskGroup(agg_op="or")  # type: ignore
     if isinstance(flat, dict):
-        for key, nest in flat.items():
-            if key not in roots:
-                continue
+        for key, nest in filter(lambda item: item[0] in roots, flat.items()):
             branch[key] = _make_tree(nest, roots, hard, desc, adj)
             branch[key]["latent"] = (  # type: ignore
                 branch[key].data != desc[key].data
@@ -674,8 +783,8 @@ def _make_tree(
 
 
 def _leaf_mask_iter(
-    branch: ty.Union[gcl.MaskGroup, gcl.MaskArray],
     branch_name: str,
+    branch: ty.Union[gcl.MaskGroup, gcl.MaskArray],
     exclude_latent: bool = True,
 ) -> ty.Generator[ty.Tuple[str, gcl.MaskArray], None, None]:
     """Recursive function, traversing a branch of the nested mask tree
@@ -688,7 +797,7 @@ def _leaf_mask_iter(
             if exclude_latent and name == "latent":
                 continue
             # TODO: look into contravariant type for this
-            yield from _leaf_mask_iter(mask, name)  # type: ignore
+            yield from _leaf_mask_iter(name, mask, exclude_latent)  # type: ignore
 
 
 def leaf_masks(mask_tree: gcl.MaskGroup) -> gcl.MaskGroup:
@@ -712,7 +821,7 @@ def leaf_masks(mask_tree: gcl.MaskGroup) -> gcl.MaskGroup:
     """
     mask_group = gcl.MaskGroup(agg_op="or")  # type: ignore
     for name, branch in mask_tree.items():
-        mask_group.update(dict(_leaf_mask_iter(branch, name)))  # type: ignore
+        mask_group.update(dict(_leaf_mask_iter(name, branch)))  # type: ignore
     return mask_group
 
 
